/*
 * Client_Admin.cpp
 *
 *  Created on: Jan 13, 2015
 *      Author: David Drake
 *
 *      This code was to separate all tasks associated with sending messages to and from the server
 *      This makes it easier to debug
 */
#include <Energia.h>
#include "blinds.h"

int parse_int(unsigned char *x){
	//take a pointer to the message buffer. Convert it and the next byte into an integer
	int y=((int)(*x)&0xFF)*256+((int)*(x+1)&0xff);
	return y;
}

//this is just a dummy function that the XBee can respond to. It will relay the message to
//the right object
void srv_message_relay(unsigned char *resp, int length){
	(*Object_Admin::Client_Admin_Ref).recv_srv_pckt(resp,length);
}

Client_Admin::Client_Admin(){

	//assume by default that we are not online, If the SD card reads network info,
	//this state will be changed
	_clt_state=CLT_OFFLINE;
	_clt_counter=0;

	Object_Admin::Client_Admin_Ref=this;
}

void Client_Admin::init(){
	//if the XBee has successfully been programmed, then bootup will work
	bootup();
}

void Client_Admin::set_online(){
	//if it is being activated to turn online, we must follow a bootup sequence
	_clt_state=CLT_IDLE;
}

void Client_Admin::set_offline(){
	//if it is being activated to turn online, we must follow a bootup sequence
	_clt_state=CLT_OFFLINE;
}

void Client_Admin::bootup(){
	/*only run if we have been explicitly turned on,
	 * this has been done to minimize wasted ressources (no need
	 * to continually try to connect if we have no network or password)
	 * the advantage to this is that we can save power
	 */

	//NOTE, this function will automatically overide any other message transfer in progress
	//that should be fine since it is always the first message to start
	if(_clt_state){
		//unlock and overide any current LED operation to lock on the purple colour

		_clt_state=CLT_BOOTUP;
		_clt_counter=5;
		send_srv_msg(_clt_state);
	}
}

int Client_Admin::get_clt_state(){
	return _clt_state;
}

void Client_Admin::send_srv_msg(byte srv_msg_type){
	byte length;

	//for all server message, turn on LEDs to show that message is being sent

	(*Object_Admin::LED_Driver_Ref).unlock();
	(*Object_Admin::LED_Driver_Ref).solid_good();
	(*Object_Admin::LED_Driver_Ref).lock();


	//currently, use short_buffer to formulate all messages except when we send the server our message.
	//In that case, send the long buffer. This allows the server to send us a new schedule
	if(srv_msg_type!=CLT_SEND_SCHD){
		length=_prep_srv_msg(Object_Admin::short_buffer, srv_msg_type);
		(*Object_Admin::XBEE_Driver_Ref).send_message(Object_Admin::short_buffer,Object_Admin::long_buffer,length,&srv_message_relay);
	}
	else
	{
		length=_prep_srv_msg(Object_Admin::long_buffer, srv_msg_type);
		(*Object_Admin::XBEE_Driver_Ref).send_message(Object_Admin::long_buffer,Object_Admin::short_buffer,length,&srv_message_relay);
	}
}

byte Client_Admin::_prep_srv_msg(unsigned char* buffer, byte srv_msg_type){
	int i=0;
	int temp_ser=SER_NUM;

	int loc_time=(*Object_Admin::Schedule_Admin_Ref).get_local_time();

	//all messages begin the same
	*(buffer+SER_INDEX_1)=(unsigned char)(temp_ser>>8)&0xff;
	*(buffer+SER_INDEX_2)=temp_ser&0xff;
	*(buffer+REQ_INDEX)=srv_msg_type;
	*(buffer+BLIND_HEIGHT_INDEX)=(*Object_Admin::Motor_Driver_Ref).get_prct_height();
	*(buffer+CT_INDEX_1)=(unsigned char)(loc_time>>8)&0xff;
	*(buffer+CT_INDEX_2)=(unsigned char)loc_time&0x00ff;

	switch(srv_msg_type){
	//for send schedule, read the entire schedule
	case CLT_SEND_SCHD:
		(*Object_Admin::Schedule_Admin_Ref).read_schedule(buffer+SCHEDULE_INDEX,128);
		i=SCHEDULE_INDEX+128;
		break;

	//for all others read the most recent update
	default:
		(*Object_Admin::Schedule_Admin_Ref).read_schedule(buffer+SCHEDULE_INDEX,8);
		i=SCHEDULE_INDEX+8;
		break;
	}
	//at this point the buffer has been prepared and is ready to send to server
	//return i as a measure of the number of bits that need to be sent
	return i;
}

void Client_Admin::recv_srv_pckt(unsigned char *resp, int length){
	int temp=0;
	//verify that this packet is destined to this machine
	//also this value will be set to 0 if the XBee object had an error sending the message
	if(length==-1||(parse_int(resp+SER_INDEX_1)!=SER_NUM)){
		_srv_pckt_failure();
	}else{
		//find actual time
		temp=parse_int(resp+CT_INDEX_1);



		if(temp!=(*Object_Admin::Schedule_Admin_Ref).get_local_time()){
					//update the clock with actual time
					(*Object_Admin::Schedule_Admin_Ref).restart_rtc_clock(temp);
		}

		//if we have just sent a bootup message, the server may have sent a message requesting our schedule

		if(*(resp+SRV_REQ_INDEX)==CLT_SEND_SCHD){
			_clt_state=CLT_SEND_SCHD;
			_clt_counter=5;


			send_srv_msg(CLT_SEND_SCHD);
		}else if(*(resp+SRV_REQ_INDEX)==CLT_RCV_SCHD&&length>=70){
			//if we requested a new schedule and the server has provided us with one, let's update
			(*Object_Admin::Schedule_Admin_Ref).write_schedule((resp+6),temp);
			_clt_state=CLT_IDLE;
			_clt_counter=0;



			(*Object_Admin::LED_Driver_Ref).unlock();
			(*Object_Admin::LED_Driver_Ref).off();


			//LEDs are locked to orange during startup sequence

		}else{
			//no new schedule, just a time update, set server state to idle
			_clt_state=CLT_IDLE;
			_clt_counter=0;


			(*Object_Admin::LED_Driver_Ref).unlock();
			(*Object_Admin::LED_Driver_Ref).off();


		}
	}
}

void Client_Admin::_srv_pckt_failure(){
	_clt_counter--;
	if(!_clt_counter){
		_clt_state=CLT_ERROR;
		//don't bother continuing this endless cycle, schedule another ping in a few minutes to an hour
		(*Object_Admin::Schedule_Admin_Ref).client_error_schd(Object_Admin::long_buffer);



		(*Object_Admin::LED_Driver_Ref).unlock();
		(*Object_Admin::LED_Driver_Ref).off();
		(*Object_Admin::LED_Driver_Ref).flash_bad();
		(*Object_Admin::LED_Driver_Ref).lock();



	}else{
		//otherwise retry sending the server the message
		send_srv_msg(_clt_state);
	}
}

void Client_Admin::check_server(){
	if(_clt_state==CLT_IDLE){
		//normally, we send our usual ping
		_clt_state=CLT_RCV_SCHD;
		_clt_counter=5;
		send_srv_msg(_clt_state);
	}
	else if (_clt_state==CLT_ERROR){
		//if we're in the error state, try to bootup
		bootup();
	}
}

void Client_Admin::check_flags(){
	/*
	if(_clt_state==CLT_IDLE){
		(*Object_Admin::LED_Driver_Ref).solid_good();
	}
	else
	{
		(*Object_Admin::LED_Driver_Ref).off();
	}
	*/
}


