/*
 * MotorDriver.cpp
 *
 *  Created on: Nov 11, 2014
 *      Author:David Drake
 *
 this file will eventually be used to give inputs to control the motor
 */


//NOTE right_handed is set to -1 by default to force the blinds to not turn if the SD card hasn't been read yet.

#include <Energia.h>
#include "Blinds.h"

 extern volatile int switch_flag=SWITCH_IDLE;
 extern volatile int enc_flag=0;

/*___________________________________ISR________________________________*/
//initialize the interupt service functions here, they will set flags.  Object_Admin will
//operate within an imperitive loop calling these objects individually to check if these variables
//have been modified
void switch_interupt()
{
	if(switch_flag==SWITCH_IDLE){
		//of this is an isolated push (not a double push), set it as such
		switch_flag=SWITCH_PUSH_1;
	}else if(switch_flag==SWITCH_COOLDOWN){
		//If a second push has been registered after the motor has started to move,
		//indicate that it is a double push
		switch_flag=SWITCH_PUSH_AGAIN;
	}else if (switch_flag==SWITCH_LOCKED){
		switch_flag=SWITCH_LOCKED_EVENT;
	}
}

void enc_interupt()
{
	//increment this interupt in case multiple interrupts occur before it can be processed
	enc_flag++;
}

/*_____________________________________Constructor and Initializer_________________________________*/
Motor_Driver::Motor_Driver()
{
	//initialize state to 0, indicating that the blinds are not moving
	_state=0;

	//set all outputs as output
	pinMode(MOT_DIR,OUTPUT);
	pinMode(MOT_ENB,OUTPUT);
	pinMode(HV_ENB,OUTPUT);

	//write all outputs to 0, this is their default state
	digitalWrite(18,HIGH);
	digitalWrite(MOT_DIR,LOW);
	digitalWrite(MOT_ENB,HIGH);
	digitalWrite(HV_ENB,LOW);

	//set all inputs as input
	pinMode(ENCA,INPUT);
	pinMode(SW1,INPUT);
	pinMode(SW2,INPUT);

	//give default values to radius and height (this is necessary since when the sd object
	//updates radius and height one at a time and on updae, both variables are needed
	//in update_enc function.
	_radius=DEFAULT_RADIUS;
	_height=DEFAULT_HEIGHT;
	update_enc();

	//cooldown reference is used to keep track of the last time a button was pressed in case a double click is detected
	//a double click tells the blinds to move all the way up or all the way down and can only be reset at the completion
	//of this motion or if the direction is reversed
	_cooldown_ref=0;

	//set right hand to -1 so that the blinds will not go up or down erroneously until
	//properly configured
	_right_handed=-1;

	//initialize enc_cooldown to 0
	_enc_cooldown=0;

	Object_Admin::Motor_Driver_Ref=this;


}

void Motor_Driver::init()
{
	//don't allow interrupt input until all objects are initialized by Object_Admin
	attachInterrupt(ENCA,enc_interupt,RISING);
	attachInterrupt(SW1,switch_interupt,CHANGE);
	attachInterrupt(SW2,switch_interupt,CHANGE);
}

void Motor_Driver::update_enc()
{
	//Each set of blinds may have a different top rail radius and length of blind material
	//this function configures Motor_Driver's height characteristic depending on the input,
	//in this way, it can be programmed for variety of different cuts and top rails

	//any dimensions for r and l can be used as long as they match (the default values are in mm)

	//the coefficient 7973 is equal to a rounded version of 360*(139+184/1221)/2pi
	_max_enc_clicks=0;//zero max clicks
	_max_enc_clicks=_height;//import height of blinds
	_max_enc_clicks=_max_enc_clicks*7973/_radius;//manipulate height, ensure divisions are done last
	_enc_clicks=_max_enc_clicks+1;//assume we are starting from top for demo purposes
}

void Motor_Driver::set_radius(int radius)
{
	_radius=radius;
	update_enc();
}

void Motor_Driver::set_height(int height)
{
	_height=height;
	update_enc();
}

void Motor_Driver::set_right_hand(short int tf)
{
	byte temp;

	if (tf)
	{
		_right_handed=1;
	}
	else
	{
		_right_handed=0;
	}
}

int Motor_Driver::get_prct_height(){
	//return percent high, recall that at _enc_clicks==_max_enc_clicks, we are are fully raised
	return _enc_clicks/(_max_enc_clicks/100);
}

void Motor_Driver::motor_up(int priority)
{
	//include support for priority. This means if a user gives input while the blinds are
	//responding to a scheduled state, the user input will overrule it.  Also check
	//that the blinds aren't already at the top. Also check that the SD card has been inserted
	//and the blinds will operate properly


	//if we're close to the top but not quite there, don't allow it to raise. The blinds will always register a few encoder clicks of
	//blinds lowering immediately after the blinds are turned off. If I don't include the factor of -500, there is a chance that the
	//blinds will always allow the user to press the up button and will turn on the blinds for a fraction of a second.
	if((_enc_clicks<_max_enc_clicks-500)&&(_right_handed>-1)&&(priority>=abs(_state)))
	{
			_state=priority;//positive state indicates blinds are moving up

			digitalWrite(HV_ENB,HIGH);
			digitalWrite(MOT_ENB,LOW);
			digitalWrite(MOT_DIR,_right_handed);

			(*Object_Admin::LED_Driver_Ref).solid_good();

	}

}

void Motor_Driver::motor_down(int priority)
{
	//include support for priority. This means if a user gives input while the blinds are
		//responding to a scheduled state, the user input will overrule it.  Also check
		//that the blinds aren't already at the top

	//+500 for the same reason as motor up
	if((priority>=abs(_state))&&(_enc_clicks>0+500)&&(_right_handed>-1))
		{
			_state=-1*priority;//negative state indicates blinds are moving down
			//analogWrite(MOT_ENB,150);
			digitalWrite(HV_ENB,HIGH);
			digitalWrite(MOT_ENB,LOW);
			digitalWrite(MOT_DIR,(!_right_handed));

			(*Object_Admin::LED_Driver_Ref).solid_good();
		}
}

void Motor_Driver::motor_stop()
{
	//ORIGINAL MOTOR_STOP


/*
	digitalWrite(MOT_ENB,HIGH);
	digitalWrite(HV_ENB,LOW);
	digitalWrite(MOT_DIR,LOW);

	(*Object_Admin::LED_Driver_Ref).off();

	//never check state, always stop when told
	_state=0;

	//we only ever get this far if we have manual stopping from the user or the encoder tells us we hav reached the end
	//this is the easiest place to ensure the switch is set to idle
	switch_flag=SWITCH_IDLE;

*/

	//newer motor stop with delayed encoder on

	//never check state, always stop when told
	_state=0;

	digitalWrite(MOT_ENB,HIGH);


	//we only ever get this far if we have manual stopping from the user or the encoder tells us we hav reached the end
	//this is the easiest place to ensure the switch is set to idle
	switch_flag=SWITCH_IDLE;
	_enc_cooldown=ENC_COOLDOWN_MAX;


}

void Motor_Driver::handle_switch(){
	volatile int val1;
	volatile int val2;

	//SW1 and SW2 are held high by default. The way this code is written indicates that if SW2 is pressed, the motor will go up
	//SW2=up, SW1=down

	if(switch_flag==SWITCH_PUSH_1)//First of potentially many events to occur, start moving or turn off motor as appropriate
	{
		//allow some time for input to debounce
		__delay_cycles(10);
		val1=digitalRead(SW1);
		val2=digitalRead(SW2);
		if(val1==val2)
		{
			motor_stop();
			//reset flag
			switch_flag=SWITCH_IDLE;
		}
		else if (val1>0)
		{
			//direct use input takes precedence over schedule
			motor_up(10);
			//keep track of the time at which this event started so that we can catch a double click
			switch_flag=SWITCH_COOLDOWN;
			_cooldown_ref=(*Object_Admin::Schedule_Admin_Ref).get_ticks();
		}
		else
		{
			//direct user input takes precedence over schedule
			motor_down(10);
			//keep track of the time at which this event started so that we can catch a double click
			switch_flag=SWITCH_COOLDOWN;
			_cooldown_ref=(*Object_Admin::Schedule_Admin_Ref).get_ticks();
		}

	}
	else if(switch_flag==SWITCH_COOLDOWN){
		val1=(*Object_Admin::Schedule_Admin_Ref).get_ticks();
		if (_cooldown_ref+COOLDOWN_PERIOD>TICKS){
			if (val1>_cooldown_ref+COOLDOWN_PERIOD-TICKS){
				//we have reached the end of our cooldown period, go into the normal loop by setting the flag to idle
				switch_flag=SWITCH_IDLE;
				//check if the button press is already removed, if so turn off the motor right away
				val1=digitalRead(SW1);
				val2=digitalRead(SW2);
				if(val1==val2)
					{
						motor_stop();
					}
			}
		}else{
			if (val1>_cooldown_ref+COOLDOWN_PERIOD){
				//we have reached the end of our cooldown period, go into the normal loop by setting the flag to idle
				switch_flag=SWITCH_IDLE;
				//check if the button press is already removed, if so turn off the motor right away
				val1=digitalRead(SW1);
				val2=digitalRead(SW2);
				if(val1==val2)
					{
						motor_stop();
					}
			}
		}
	}
	else if (switch_flag==SWITCH_PUSH_AGAIN){
		//if we've registered a second event within half a second:

			//allow some time for input to debounce
			__delay_cycles(10);
			val1=digitalRead(SW1);
			val2=digitalRead(SW2);
			if(val1==val2)
			{
				//This will happen if a double click occurs (in between there will be a temporary equal state)
				//make sure the state returns to switch cooldown
				switch_flag=SWITCH_COOLDOWN;
			}
			else if ((val1>0&&_state>0)||(val2>0&&_state<0))
			{
				//lock upward/downward motion by changing the flag
				switch_flag=SWITCH_LOCKED;
			}
			else
			{
				//reverse direction of motor
				if (val1>0){
					motor_up(10);
				} else {
					motor_down(10);
				}
				//perform the cooldown
				switch_flag=SWITCH_COOLDOWN;
				_cooldown_ref=(*Object_Admin::Schedule_Admin_Ref).get_ticks();
			}
	}
	else if (switch_flag==SWITCH_LOCKED_EVENT){
		//if the opposite button is pressed, allow the blinds to move in the opposite direction.
		//treat this as a new event, put the state into cooldown
		val1=digitalRead(SW1);
		val2=digitalRead(SW2);
		if (_state==0){
			motor_stop();
			switch_flag=SWITCH_IDLE;
		}

		if((val2==0&&_state<=0)||(val1==0&&_state>=0)){
			//reverse direction of motor
			if (val1>0){
				motor_up(10);
			} else {
				motor_down(10);
			}
			//perform the cooldown
			switch_flag=SWITCH_COOLDOWN;
			_cooldown_ref=(*Object_Admin::Schedule_Admin_Ref).get_ticks();
		}else{
			//return to locked state
			switch_flag=SWITCH_LOCKED;
		}
	}


}

void Motor_Driver::handle_enc()
{
	/*this is a quadrature encoder, there are two channels A and B (encA, encB)
	 * which produce two step wave forms that are out of sync by 90 degrees. They are triggered
	 * 360 times per revolution of the DC motor (not the gear box).  Every turn corresponds to
	 * 50+801/895 turns of the DC motor or 18322 encoder clicks per revolution (crazy huge number!)
	 * It is not feasible to trigger every event, if we consider both channels (*2 interrupts) and
	 * both rising and falling edges (*2 interrupts) we would have a 4x multiplier which is rediculous.
	 * We will trigger on EncA only.  We could determine direction of travel by comparing EncA to EncB.
	 * For example if EncA is leading by 90 degrees (indicating one direction of motion) it will never
	 * equal EncB (at the time of interrupt). If EncA is trailing by 90 degrees, it is travelling in the opposite direction and
	 * it will always equal EncB (at the time of interrupt). This is overkill for this application.
	 * We will turn on the motor and it will turn in a known direction, so we will just increment or
	 * decrement as appropriate
	 */


	if(enc_flag>0)
	{
		if(_state>0)
		{
			//add as many clicks as have occured recently
			_enc_clicks+=enc_flag;
			//reset the flag immediately
			enc_flag=0;
			//if we have almost reached the top, stop
			if(_enc_clicks>=_max_enc_clicks-250)
			{
				motor_stop();
			}
		}
		else if ((_state<=0))
		{
			//always assume if we aren't rising, encoder clicks are contributing to the blinds going down (because the blinds will slide slightly with friction brake)
			//subtrack according to the number of events that have occured
			_enc_clicks-=enc_flag;
			//reset the flag immediately
			enc_flag=0;
			//if we have almost reached the bottom, stop
			if(_enc_clicks<0+250)
				{
					motor_stop();
				}
		}


	}else if(_enc_cooldown!=0){
		//if we have recently turned off the motor state will be 0 but we will have a value for encoder cooldown, HV is still on so that the encoder is turned on
		_enc_cooldown--;
		//cooldown will count down a number of MSP loops ONLY IF no encoder clicks are recorded this round. This will guarentee that
		//all blind slippage will be recorded.
		if(_enc_cooldown==0){
			//if we have finished the encoder cooldown, turn off all other motor controls.
			digitalWrite(HV_ENB,LOW);
			digitalWrite(MOT_DIR,LOW);
			(*Object_Admin::LED_Driver_Ref).off();
		}
	}
}
void Motor_Driver::check_flags()
{
	handle_switch();
	handle_enc();

	if(_state!=0){
		//overwrite LEDs to ensure that it shows that the motor is moving
		(*Object_Admin::LED_Driver_Ref).solid_good();

	}
}


