/*
 * SD_Driver.cpp
 *
 *  Created on: Nov 24, 2014
 *      Author: David Drake
 */

#include <Energia.h>
#include "SPI.h"
#include "pfatfs.h"
#include "Blinds.h"

extern volatile short sd_flag=0;

/*____________________________Generic processing and interrupt functions_____________________*/

void cs_handler()
{
	sd_flag=1;
}

void timeout(){
	//watchdog will restart the system
	WDTCTL = WDTPW;
	while(1){}
}

SD_Driver::SD_Driver()
{
	//set cspin to an input
	pinMode(SD_CS,INPUT);

	Object_Admin::SD_Driver_Ref=this;
}

void SD_Driver::init()
{
	if(digitalRead(SD_CS))
	{
		read_SD();
		listen();
	}
	else
	{
		//allow circuit to restart and timeout, this will continue until the SD card is read succesfully
		timeout();

	}
}

void SD_Driver::read_SD()
{
	int _temp_int;
	int _index;

	//this function will read the SD and program the other objects accordingly

	volatile int rc;
	unsigned short br;
	char *rx_buf=(char *)Object_Admin::long_buffer;//use the larger buffer
	char *tx_buf=(char *)Object_Admin::short_buffer;//use the smaller buffer
	unsigned int local_timestamp;

	//turn on both colour of LED (good LED) to indicate that SD is being read
	(*Object_Admin::LED_Driver_Ref).solid_all();


	__delay_cycles(1000);
	FatFs.begin(SD_CS);
	__delay_cycles(100);
	rc = FatFs.open(FILE);
	__delay_cycles(100);

	//read the blind length
	rc = FatFs.read(rx_buf, 6, &br);


	if(rc!=0){
		(*Object_Admin::LED_Driver_Ref).off();
		(*Object_Admin::LED_Driver_Ref).solid_bad();
		timeout();
	}

	_index=_find_start((unsigned char *)rx_buf,br);


	_temp_int=_parse_int((unsigned char *)rx_buf,(_index),5);


	//set blind length details
	(*Object_Admin::Motor_Driver_Ref).set_height(_temp_int);

	//read the top rail diameter
	rc = FatFs.read(rx_buf, 6, &br);


	if(rc!=0){
		(*Object_Admin::LED_Driver_Ref).off();
		(*Object_Admin::LED_Driver_Ref).solid_bad();
		timeout();
	}

	_index=_find_start((unsigned char *)rx_buf,br);
	_temp_int=_parse_int((unsigned char *)rx_buf,(_index),5);
	_temp_int=_temp_int/2;//the java wizard saves diameter, but we want radius

	//set motor details
	(*Object_Admin::Motor_Driver_Ref).set_radius(_temp_int);

	//read if it is right handed
	rc = FatFs.read(rx_buf, 6, &br);
	//read the top rail diameter

	if(rc!=0){
			(*Object_Admin::LED_Driver_Ref).off();
			(*Object_Admin::LED_Driver_Ref).solid_bad();
			timeout();
	}

	_index=_find_start((unsigned char *)rx_buf,br);
	_temp_int=(unsigned short)_parse_int((unsigned char *)rx_buf,(_index),5);

	//set motor to right handed or not
	(*Object_Admin::Motor_Driver_Ref).set_right_hand(_temp_int);

	//read network name
	rc = FatFs.read(rx_buf, PWD_NETWORK_LEN, &br);
	//read the top rail diameter



	if(rc!=0){
			(*Object_Admin::LED_Driver_Ref).off();
			(*Object_Admin::LED_Driver_Ref).solid_bad();
			timeout();
	}

	_index=_find_start((unsigned char *)rx_buf,br);

	//if no name present, than xbee will be set offline
	//same goes for client admin (this will result in power savings so we don't
	//needlessly try connecting to the internet
	if(_index==PWD_NETWORK_LEN){
		(*Object_Admin::XBEE_Driver_Ref).set_offline();
		(*Object_Admin::Client_Admin_Ref).set_offline();
	}
	else
	{

		//both SD card and XBee communicate using SPI. This means that we must deactivate the SPI line
		//to the SD card and give Xbee_driver control to send a message.
		//digitalWrite(SD_CS,HIGH);
		(*Object_Admin::XBEE_Driver_Ref).set_network((unsigned char*)(rx_buf+_index),(unsigned char*)tx_buf,PWD_NETWORK_LEN-_index);

		digitalWrite(SD_CS,LOW);
		(*Object_Admin::Client_Admin_Ref).set_online();
	}

	//read network password
	rc = FatFs.read(rx_buf, PWD_NETWORK_LEN, &br);


	if(rc!=0){
			(*Object_Admin::LED_Driver_Ref).off();
			(*Object_Admin::LED_Driver_Ref).solid_bad();
			timeout();
	}

	_index=_find_start((unsigned char *)rx_buf,br);
	//double check to make sure that we should be connecting to the internet. This is something
	//Schedule_Admin keeps track of since it controls operation of XBEE.  Do not check
	//if Xbee is connected as it may have trouble connecting to the network
	if((*Object_Admin::Client_Admin_Ref).get_clt_state()==CLT_IDLE)
	{
		(*Object_Admin::XBEE_Driver_Ref).set_password((unsigned char *)(rx_buf+_index),(unsigned char*)tx_buf,PWD_NETWORK_LEN-_index);
	}

	//from now on, if the reads fail, that is no longer a concern we will continue with error coding and timing out.

	//read the timestamp info. Use appropriate bit masks
	rc = FatFs.read(rx_buf, 2, &br);


	local_timestamp=((unsigned int)rx_buf[0]&0xff)*256+((unsigned int)rx_buf[1]&0xff);

	//read the schedule info
	rc = FatFs.read(rx_buf, 64, &br);


	rc = FatFs.read(rx_buf+64, 64, &br);
	(*Object_Admin::Schedule_Admin_Ref).write_schedule((unsigned char *)rx_buf,local_timestamp);



	digitalWrite(18,LOW);

	FatFs.close();

	//message that SD is done by turning off LEDs
	(*Object_Admin::LED_Driver_Ref).off();




}

void SD_Driver::listen()
{
	//whenever values are read from the SD card, cs is set as an output. This function here will
	//set it as an input and wait for a change. In the even that the SD card is added or removed,
	//the circuit will restart.  It will only ever operate if an SD card with valid information is
	//inserted
	pinMode(SD_CS,INPUT);
	attachInterrupt(SD_CS,cs_handler,CHANGE);
}

void SD_Driver::stop_listen()
{
	//turn off the interupt, the program may get confused when reading the SD card since SD_CS
	//gets turned into an output
	detachInterrupt(SD_CS);
}

void SD_Driver::check_flags()
{
	if (sd_flag==1)
	{
		//we're going to reset the system whenever the SD card is added or removed.
		//Only if the SD card is inserted will the blinds operate (simple way of deactivating the blinds)

		//stop the motor if it's moving
		(*Object_Admin::Motor_Driver_Ref).motor_stop();

		//watchdog will restart the system
		timeout();
	}
}
/*______________________Private Methods___________________________________*/
int SD_Driver::_find_start (unsigned char *buf, int msg_len)
{
	//search through the message, find the end of the padding
	//return the index of the buffer that corresponds to that
	int i;

	for ( i=0 ; i < msg_len; i++){
	   if(*(buf+i)!=(char)0){
		   //we've found the start
		   return i;
		   break;
	   }
	}
	return msg_len;
}

int SD_Driver::_parse_int (unsigned char* buffer,int start_index,int end_index)
{
	//start at the beginning of the string representation of a number
	//add each integer and multiply it by its appropriate power
	//this assumes that the buffer passed ONLY contains chars that represent integers
	int i=start_index;
	int total=0;

	for(i=start_index;i<=end_index;i++)
	{
		total=total*10;
		total+=(int)*(buffer+i)-48;
	}
	return total;
}


