/*
 * RTC_Clock.cpp
 *
 *  Created on: Nov 11, 2014
 *      Author: David Drake
 *
 *      This is the core of the scheduling of time keeping of the msp430
 *      Schedule will be saved in flash memory, clock will be saved in normal memory as a static
 *      variable because it will need to be refreshed if the circuit loses power, and is updated
 *      by an inturupt.
 *
 *      The clock is run by static functions and variables because the ISR does not like to reference
 *      object methods. The static methods will keep track of time, and every minute they will
 *      use a static pointer to the object to check if a scheduled activity needs to happen
 *
 *      The object itself administers and protects references to the schedule. The schedule is saved
 *      in flash memory so that if power is lost, the schedule is not (if a RTC module is added, then
 *      the blinds can continue as previous). Schedule is saved in blocks C and B.
 *
 *
 *      Block C will keep overhead information and repetitive events.  It will save them in the
 *      following format: Unix timestamp of last update (2 words), local timestamp (mins since sunday)
 *      (1 word) reserved word for future use followed by repetitive events.  Repetitive events
 *      take 2 words each, each word follows the standard format, the leading 2 bits represent
 *      the action that needs to happen for the event, the following 14 bits are a local timestamp
 *      (minutes since sunday at 0). Word one will have the event
 *      separation timestamp (time in between events) and Word two will have the last time the event
 *      occured.  32 words are available. 2 words for the timestamp means that 30 words are left.
 *      Therefore 15 repetitive events can be stored.  These repetitive events are intended to be used
 *      to trigger the XBee checkins and for repetitive up/downs (like everyday at 7:00 am raise blinds)
 *
 *      Block B will save specific events.  Each event gets a word, the leading two bits indicate the action
 *      while the next 14 bits are the timestamp (minutes since minute 0 on sunday).  This can hold 32 events
 *      that are not separated by any specific time (there can be an average of 4.5 updown events per day)
 *
 *      In the future, if this is not adequate, a wifi connected set of blinds can have their schedule
 *      refreshed frequently (every day, reload the schedule).
 *
 */

#include <Energia.h>
#include <MspFlash.h>
#include "blinds.h"

extern volatile unsigned int _clock_counter=0;//keeps track of the timechunks that have passed according to TA1
//set _overflowing flag when dealing with a minute event. If another minute event occurs when the
//first hasn't been dealt with, this flag will prevent them from conflicting

Schedule_Admin::Schedule_Admin()
{
	//set next_event time to an impossible value (greater than the number of minutes in a week)
	//this will signal to scheduler that no schedules will be used today
	_next_event_time=10100;
	Object_Admin::Schedule_Admin_Ref=this;
}

void Schedule_Admin::init()
{
	_find_next_event();
}

void Schedule_Admin::write_schedule(unsigned char *new_schd, int time_stamp)
{
	//once another object receives a new schedule, it will call this function to update the schedule
	//the time of the blinds must be updated everytime the schedule is updated to ensure the blinds
	//operate at a reasonable time.

	/*memory is stored according to the following convention:
	 * first 4 words unix timestamp (set by server),
	 * second 2 words reserved for internet scheduling event
	 * next 26 words are set by server for other repetitive events
	 * next 32 words are set by server for unique events
	 */

	//step one is to modify the incoming schedule and add our own repetitive event
	//for checking online (if necessary)
	switch((*Object_Admin::Client_Admin_Ref).get_clt_state()){
	case(CLT_OFFLINE):
			//placeholder repetitive item, this will never be activated
			*(new_schd+8)=0x68;
			*(new_schd+9)=0x00;
			break;

	case(CLT_ERROR):
			//check in an hour (no need to waste power if the server is down
			*(new_schd+8)=0x50;
			*(new_schd+9)=0x00;
			break;

	default:
		//for all other cases, let's check in at the 5 minute mark
			*(new_schd+8)=0x48;
			*(new_schd+9)=0x00;
			break;
	}

	//write the first half of the schedule (unix timestamp and repetitive memory)
	Flash.erase(REPETITIVE_MEM);
	Flash.write(REPETITIVE_MEM,new_schd,64);

	//write the second half of the schedule (individual memory)
	Flash.erase(NON_REPETITIVE_MEM);
	Flash.write(NON_REPETITIVE_MEM,(new_schd+64),64);

	//update the current clock
	restart_rtc_clock(time_stamp);
}

void Schedule_Admin::client_error_schd(unsigned char *buf)
{
	//if the client admin has an error sending/receiving a message, it will call this function.
	//we won't try reconnecting for an hour, and will keep that schedule until we have a successful connection

	//read the current repetitive schedule
	Flash.read(REPETITIVE_MEM, buf, 64);

	//replace the xbee checkup to make it once every hour
	*(buf+8)=0x50;
	*(buf+9)=0x00;

	//write the first half of the schedule (unix timestamp and repetitive memory)
	Flash.erase(REPETITIVE_MEM);
	Flash.write(REPETITIVE_MEM,buf,64);

	//update the current clock
	_find_next_event();
}

void Schedule_Admin::client_normal_schd(unsigned char *buf)
{

	//read the current repetitive schedule
	Flash.read(REPETITIVE_MEM, buf, 64);

	//replace the xbee checkup (always do this to ensure circuit stays up to date)
	*(buf+8)=0x48;
	*(buf+9)=0x00;

	//write the first half of the schedule (unix timestamp and repetitive memory)
	Flash.erase(REPETITIVE_MEM);
	Flash.write(REPETITIVE_MEM,buf,64);

	//update the current clock
	_find_next_event();
}

void Schedule_Admin::read_schedule(unsigned char *buf,int num_bytes){
	//this will write the memory directly into the provided buffer
	Flash.read(REPETITIVE_MEM, buf, num_bytes);
}

int Schedule_Admin::get_local_time(){
	return _rtc_clock;
}

void Schedule_Admin::restart_rtc_clock(int new_time)
{
	//This function sets all the parameters for the static functions
	//ensure all static variables are initialized using the timestamp parameter
	//Set the _ptr to this, be careful and don't instantiate a second Schedule_Admin object
	_rtc_clock=new_time;
	_clock_counter=0;

	//Set timer options.  These predefined values come from wiring and are defined in the header file

	TA1CCR0 = F_CPU/490;           // This matches the PWM period used for analogue write, capture and compare register
	//CCR0 is the maximum value that the internal counter goes to
	TA1CTL = TASSEL_2 + MC_1;       // SMCLK, up mode, this is the control register. SMCLK is sourced from DCOCLK
	TA1CCTL0 |= CCIE;				//capture and compare control register

	//find the next event so that scheduled operation can begin as it is supposed to
	_find_next_event();
}

void Schedule_Admin::_check_schedule()
{
	if(Schedule_Admin::_rtc_clock==_next_event_time)
	{
		switch(Schedule_Admin::_next_event_type)
		{
		case 0xC0:
			(*Object_Admin::Motor_Driver_Ref).motor_up(2);
			break;

		case 0x80:
			(*Object_Admin::Motor_Driver_Ref).motor_down(2);
			break;

		case 0x40:
			(*Object_Admin::Client_Admin_Ref).check_server();
			break;
		}
		//there is an extra control bit sequence that can be used in future.
		_find_next_event();
	}
}

int Schedule_Admin::_schedule_exists()
{
	//this is a very basic function. Check if a unix timestamp exists. In the future,
	//a checksum could be included as well
	unsigned char mem_cpy[WORD_REF*2]={};
	Flash.read(REPETITIVE_MEM, mem_cpy, WORD_REF*2);
	//this compiler likes to analyze it per byte, check that all 4 bytes in the unix timestamp
	//are not 0xFF
	for(int i=0;i<WORD_REF*2;i++)
	{
		if(mem_cpy[i]!=0xFF)
			{
				return 1;
			}
	}
	return 0;
}

unsigned int Schedule_Admin::_rep_mem_ptr_to_timestamp(unsigned char* mem_cpy)
{
	unsigned int current_time=Schedule_Admin::_rtc_clock;

	//the compiler is not cooperating, we are using pointers in this function to force it
	//to allocate memory that can then be manipulated to decipher the repetitive schedule item
	unsigned int rep_type=(*mem_cpy&0x38);

	//referencing the char pointer allows us to access one byte at a time, this
	//expression accesses and weighs the memory locations according to the
	//method developed for this circuit
	unsigned int offset=(*mem_cpy&0x07)*256+*(mem_cpy+1);
	unsigned int x=1;
	unsigned int y=current_time;

	unsigned int ans;
	//analyze the type of repetition (found in rep_type using a bit mask). Currently only supporting
	//the following: 0x3800->daily, 0x3000->week days, 0x1000->every ten minutes,
	//0x0800->every five minutes and 0x0000->every minute



	//REPETITIVE EVENTS WILL NOT WORK ON SUNDAYS AT MIDNIGHT WHEN THE CLOCK ROLLS OVER
	//USE THE SERVER TO PROGRAM THIS AT 11:59 TO GET AROUND THIS
	switch(rep_type)
	{
	case 0x38:
			x=1440;
			break;

	case 0x30:
			if(Schedule_Admin::_rtc_clock>8640)
			{
				/*check if we're on the weekend, if time is greater than 8640, its saturday,
				 * we don't need to worry about the next event until next week, so return an
				 * impossibly large value
				 */
				return 10100;
			}
			else if (Schedule_Admin::_rtc_clock<1440)
			{
				//check if we're on sunday, return the first event time on monday
				return 1440+offset;
			}
			else
			{
				//we got here because we're in the regular week
				x=1440;
			}
		break;
	case 0x28:
		//reserve this for NO ACTION (IE return an impossible value)
		x=10100;
	case 0x10:
		//hourly
			x=60;
			break;
	case 0x8:
			x=5;
			break;

	case 0x0:
			x=1;
			break;
	}

	if(x==1&&offset==0){
		//just return the max possible value if we have an unreasonable repetitive memory event
		return 10081;
	}

	y=offset;

	while (y<10081&&y<=(current_time-offset)){
		y+=x;
	}
	__delay_cycles(10);

	return  y;
}

unsigned int Schedule_Admin::_mem_ptr_to_timestamp(unsigned char mem_cpy[WORD_REF])
{
	int next_time=(mem_cpy[0]&0x3f)*256+mem_cpy[1];
	int x=Schedule_Admin::_rtc_clock;
	if (next_time<=x&&x!=0)
	{
		//this event has already happened this week, return it offset by a week (this is an impossible
		//value and of little consequence.  Add an exception for x=0, this algorithm is designed for
		//to look at the next event from an earlier time. We do not analyze time from the negative
		//side (ie x=-1) therefore this exception must be included
		next_time=next_time+10080;
	}
	return next_time;
}

//private functions
void Schedule_Admin::_find_next_event()
{
	//check to see if a schedule exists in memory. If not, set next event time to an impossible value
	//to ensure the blinds don't act by themselves.

	//allocate memory to copy from the flash memory
	unsigned char mem_cpy[WORD_REF]={};
	//initialize a pointer to the first repetitive memory location (after the unix timestamp)
	unsigned char *temp_ptr=REPETITIVE_MEM+8;

	//create local copies of next event time and type, while searching, we will modify these
	//values so that in the event that we're interrupted, we won't accidently point to the
	//wrong next event.  Use pointers since this compiler doesn't want to allocate it's own
	//ressources
	volatile unsigned int temp_next_event_time=10100;
	volatile unsigned char temp_next_event_type=0;
	volatile unsigned int temp_timestamp=0;

	if(_schedule_exists())
	{
		temp_next_event_time=10100;
		temp_next_event_type=0;

		//read the first set of memory (Flash requires number of bytes to read)

		while(temp_ptr<NON_REPETITIVE_MEM+64)
		{
			//read first memory block
			Flash.read(temp_ptr, mem_cpy, WORD_REF);

			//loop past empty memory 0x0000 and 0xffff (only analyze interesting memory locations
			//Note that 0xffff and 0x0000 are reserved and assumed as blank

			//determine the next soonest event time for this timestamp
			if(temp_ptr<NON_REPETITIVE_MEM)
			{
				temp_timestamp=_rep_mem_ptr_to_timestamp(mem_cpy);
			}
			else
			{
				temp_timestamp=_mem_ptr_to_timestamp(mem_cpy);
			}
			//if we have found an event that happens sooner replace our current temporary next timestamp
			if(temp_timestamp<temp_next_event_time)
			{
				temp_next_event_time=temp_timestamp;
				temp_next_event_type=mem_cpy[0];
			}
			else if (temp_timestamp==temp_next_event_time)
			{
				//if they're the same, determine which has order of precedence
				//this is determined by the leading bits, therefore we can directly compare
				//the two numbers

				//THIS IS SLOPPY.  If two up events occured at the same time, one arbritrary
				//action would be chosen. However, that is a silly example, and not worth the
				//effor to fix this
				if(mem_cpy[0]>temp_next_event_type)
				{
					temp_next_event_time=temp_timestamp;
					temp_next_event_type=mem_cpy[0];
				}
			}

			//we have finished evaluating this repetitive item, continue to next word
			temp_ptr=temp_ptr+WORD_REF;
		}

		//write results to schedule variables
		_next_event_time=temp_next_event_time;
		//for now, just keep the whole value since its easier than shifting bits around
		_next_event_type=temp_next_event_type&0xC0;
	}

	else
	{
		//give an impossible value for next event time, and give placeholders to other attributes
		//there is no reliable schedule
		//set next event time to a minute that exceeds a single week, an impossible value
		_next_event_time=10100;
		//initialize _next_event_type to 0, just so it has a value
		_next_event_type=0;
		//initialize the next event pointer to the first repetitive block
	}
}

// called by the ISR every time an inturrupt is received
void Schedule_Admin::_update_rtc_clock()
{
	  _rtc_clock++;

	  //check if we have reached the end of the week, and rollover rtc_clock if necessary
	  if (_rtc_clock==MINS_PER_WEEK)
	 	  {
		  // it is forbidden to assign an event to occur as MINS_PER_WEEK.
		  //instead give it a time of 0, find_next_event can find events that occur at t=0
	 		  _rtc_clock=0;
	 		 _find_next_event();
	 	  }
	  //check the schedule to see if anything should be happening now
	  _check_schedule();

	  //testing code when every minute it updates it's schedule to see if a new event is happening.
	  _find_next_event();
}

int Schedule_Admin::get_ticks()
{
	//This function will return the current tick count held in _clock_counter
	//it is intended to be used by the motor module so that it can determine when the blinds
	//bottom out (the time between encoder ticks will increase)
	return _clock_counter;
}

void Schedule_Admin::check_flags()
{
	if(_clock_counter>=TICKS)
	{
		//a minute has passed, perform actions
		 // Interrupts are at 490Hz, every 490 ticks, a second has passed
			//perhaps in the future, if pwm libarary is modified, we can run at 500Hz and then we can combine
			//_clock_counter and _rtc_clock

		/*The circuit operates on a minute schedule BUT the value of TICKS can be changed for
			   * debugging purposes. If we are at a new minute, restart the _clock_counter, update
			   * the rtc_clock variable and check to see if an action needs to be performed.
			   */

		_clock_counter-=TICKS;
		//reset the internal timer immediately
		_update_rtc_clock();
	}
}

// Timer interrupt service routine
__attribute__((interrupt(TIMER1_A0_VECTOR)))
void Timer_A_int(void)
{
	//update the clock
	_clock_counter++;
}

